/*				21-Mar-90 10:29:19am										*/

/*--------------------------------------------------------------------------*/
/* History: termtest.c, V0.0, 28-Feb-90 10:34:48pm, Dan Hildebrand, Baseline*/
/*----------------------------------------------------------------------------

     Test routine for the term and terminfo routines.

----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <term.h>
#include <sys/dev.h>
#include <sys/qnxterm.h>
#include <sys/mouse.h>

int delay_mode = 0, keypad_mode = 1, half_mode = 0, timeout_mode;
int mouse, resize;

#define	t	term_state
#define ct	__cur_term

void filled_box( void );
void attr_test( void );
void term_ftest( void );
void color_test( void );
void speed_test( void );
void out_test( void );
void show_status( void );
void tparm_test( void );
void pop_test( void );
void do_shell( void );
void mouse_test( void );
char *key_name( unsigned c );

unsigned text_color = TERM_CYAN | TERM_BLUE_BG;

main() {
	int y, count = 0;
	unsigned ch = 0;
	
	if ( term_load() < 0 ) {
		fprintf( stderr, "termtest: Unable to load terminal definition\n" );
		perror( "Error: " );
		exit( -1 );
		}
	term_video_on();				/* Enable terminal shadow image			*/
#ifdef NEVER
	resize = (term_resize_on() == 0);	/* Enable the resize keyboard event	*/
	mouse = (term_mouse_on() == 0);	/* Attempt to attach to the mouse too	*/
#endif
	term_color( text_color );
	term_fill( text_color );

	for(;;) {
		term_color( text_color );
		term_clear( TERM_CLS_SCR );
		term_type( y=2,  4, "QNX Term / Terminfo Test Program", 0, TERM_HILIGHT);
		term_type( y+=2, 6, "1) Input test", 0, 0 );
		term_type( ++y,  6, "2) Output test", 0, 0 );
		term_type( ++y,  6, "3) Speed test", 0, 0 );
		term_type( ++y,  6, "4) Tparm test", 0, 0 );
		term_type( ++y,  6, "5) Pop-up menu test", 0, 0 );
		term_type( ++y,  6, "6) Execute a shell", 0, 0 );
		term_type( ++y,  6, "7) Mouse test", 0, 0 );
		term_type( ++y,  6, "m) Mouse on/off", 0, 0 );
		term_type( ++y,  6, "r) Resize on/off", 0, 0 );
		term_type( ++y,  6, "x) Exit", 0, 0 );
		term_printf( 20, 0, TERM_NORMAL, "Mouse : %s", mouse ? "enabled" : "disabled" );
		term_printf( 21, 0, TERM_NORMAL, "Resize: %s", resize ? "enabled" : "disabled" );
		term_printf( 22, 0, TERM_NORMAL, "Video : %s", ct->_cc ? "enabled" : "disabled" );
		term_printf( 23, 0, TERM_NORMAL, "cp=%04x dp=%04x mp=%04x c=%d ch=%04x",
			ct->_cproxy, ct->_dproxy, ct->_mproxy, count, ch );
		term_type( y+=2, 4, "Select: ", 0, TERM_HILIGHT );

		ch = term_key();
		++count;
		switch( ch ) {
			case '1':	in_test(); 					break;
			case '2':	out_test(); 				break;
			case '3':	speed_test(); 				break;
			case '4':	tparm_test(); 				break;
			case '5':	pop_test(); 				break;
			case '6':	do_shell();					break;
			case '7':	mouse_test();				break;
			case 'm':	if ( mouse ) {
							term_mouse_off();
							mouse = 0;
							}
						else {
							mouse = (term_mouse_on() == 0);
							}
						break;
			case 'r':	if ( resize ) {
							term_resize_off();
							resize = 0;
							}
						else {
							resize = (term_resize_on() == 0);
							}
						break;
			case K_RESIZE:
						/* No processing required, term routines did it	*/
						break;
			case K_ESC:
			case '0x18':	/* ^x		*/
			case 'x':	term_cur( t.num_rows-1, 0 );
						term_restore();
						term_clear( TERM_CLS_EOS );
						exit( 0 );
			}
		}
	}

void
out_test() {
	prompt( "Clear screen and draw some cursor addressed text:" );
	filled_box();

	prompt( "At Line 6, two lines will be deleted:" );
	term_cur( t.num_rows-1, 0 );
	term_clear( TERM_CLS_EOS );
	term_delete_line( 6, 2 );

	prompt( "Three lines will be inserted before line 3:" );
	term_cur( t.num_rows-1, 0 );
	term_clear( TERM_CLS_EOS );
	term_insert_line( 3, 3 );

	prompt( "Five characters will be deleted from the start of line 2:" );
	term_cur( 2, 0 );
	term_delete_char( 5 );

	prompt( "Three spaces will be inserted on line 1:" );
	term_cur( 1, 0 );
	term_insert_char( 3 );

	prompt( "Text attributes test:" );
	attr_test();

	prompt( "Draw a bold box:" );
	term_clear( TERM_CLS_SCR );
	term_box( 3, 3, 60, 10, TERM_HILIGHT );

	prompt( "Draw a filled box:" );
	term_clear( TERM_CLS_SCR );
	term_box_fill(	3, 3, 60, 10, TERM_YELLOW | TERM_CYAN_BG | TERM_HILIGHT,
					TERM_BOX_FRAME, ' ' );

	prompt( "Draw a filled, double line box:" );
	term_clear( TERM_CLS_SCR );
	term_box_fill(	3, 3, 60, 10, TERM_YELLOW | TERM_CYAN_BG | TERM_HILIGHT,
					TERM_BOX_DOUBLE, ' ' );

	prompt( "Draw a filled, double top line box:" );
	term_clear( TERM_CLS_SCR );
	term_box_fill(	3, 3, 60, 10, TERM_YELLOW | TERM_CYAN_BG | TERM_HILIGHT,
					TERM_BOX_DBL_TOP, ' ' );

	prompt( "Draw a filled, double side line box:" );
	term_clear( TERM_CLS_SCR );
	term_box_fill(	3, 3, 60, 10, TERM_YELLOW | TERM_CYAN_BG | TERM_HILIGHT,
					TERM_BOX_DBL_SIDE, ' ' );

	prompt( "Draw a full screen box:" );
	term_clear( TERM_CLS_SCR );
	term_color( text_color );
	term_box( 0, 0, t.num_cols, t.num_rows, TERM_HILIGHT );
	if ( auto_right_margin ) {
		term_type( 12, 10, "This terminal does have the auto_right_margin capability", 0, TERM_NORMAL );
		term_type( 13, 10, "defined, so the bottom right corner should be missing, as", 0, TERM_NORMAL );
		term_type( 14, 10, "the term_box routines automatically avoids this corner on", 0, TERM_NORMAL );
		term_type( 15, 10, "this terminal type to prevent the screen from scrolling up.", 0, TERM_NORMAL );
		}
	else {
		term_type( 12, 10, "This terminal does not have the auto_right_margin capability", 0, TERM_NORMAL );
		term_type( 13, 10, "defined, so the bottom right corner should be properly", 0, TERM_NORMAL );
		term_type( 14, 10, "displayed and the screen should not have scrolled.", 0, TERM_NORMAL );
		}
	term_type( 20, 10, "Press <Enter> for the filled box test:", 0, TERM_NORMAL );
	term_key();

	term_clear( TERM_CLS_SCR );
	term_color( text_color );
	term_box_fill( 0, 0, t.num_cols, t.num_rows,
		TERM_YELLOW | TERM_CYAN_BG | TERM_HILIGHT, TERM_BOX_FRAME, ' ' );
	if ( auto_right_margin ) {
		term_type( 12, 10, "This terminal does have the auto_right_margin capability", 0, TERM_NORMAL );
		term_type( 13, 10, "defined, so the bottom right corner should be missing, as", 0, TERM_NORMAL );
		term_type( 14, 10, "the term_box routines automatically avoids this corner on", 0, TERM_NORMAL );
		term_type( 15, 10, "this terminal type to prevent the screen from scrolling up.", 0, TERM_NORMAL );
		}
	else {
		term_type( 12, 10, "This terminal does not have the auto_right_margin capability", 0, TERM_NORMAL );
		term_type( 13, 10, "defined, so the bottom right corner should be properly", 0, TERM_NORMAL );
		term_type( 14, 10, "displayed and the screen should not have scrolled.", 0, TERM_NORMAL );
		}
	term_type( 20, 10, "Press <Enter> for the left side scroll test:", 0, TERM_NORMAL );
	term_key();

	scroll_box_left();
	scroll_box_right();

	prompt( "Term_field test:" );
	term_color( text_color );
	term_ftest();

	prompt( "Color test:" );
	color_test();
	term_key();
	}

void
filled_box() {
	int r, c;
	char ch;

	term_clear( TERM_CLS_SCR );
	term_type( 0, 0, "Each character shown is individually addressed.", 0, TERM_NORMAL );
	for ( r = 1; r < 10; ++r )
		for ( c = 0; c < 70; ++c ) {
			ch = (char) (r + '0');
			term_type( r, c, &ch, 1, TERM_NORMAL );
			}
	}

void
scroll_box_left() {
	int i, j;

	if ( term_delete_char(0) < 0 ) {
		prompt( "Delete_char not supported on this terminal:" );
		return;
		}
	term_clear( TERM_CLS_SCR );
	term_color( text_color );
	term_box_fill( 0, 0, t.num_cols, t.num_rows,
		TERM_YELLOW | TERM_CYAN_BG | TERM_HILIGHT, TERM_BOX_FRAME, ' ' );
	term_type( 10, 5, "Press <enter> for the left side scroll:", 0, TERM_NORMAL );
	term_key();
	for( i = 0; i < t.num_cols; ++i ) {
		for( j = 0; j < t.num_rows; ++j ) {
			term_cur( j, 0 );
			term_delete_char( 1 );
			}
		term_flush();
		}
	}

void
scroll_box_right() {
	int i, j;

	if ( term_insert_char(0) < 0 ) {
		prompt( "Insert_char not supported on this terminal:" );
		return;
		}
	term_clear( TERM_CLS_SCR );
	term_color( text_color );
	term_box_fill( 0, 0, t.num_cols, t.num_rows,
		TERM_YELLOW | TERM_CYAN_BG | TERM_HILIGHT, TERM_BOX_FRAME, ' ' );
	term_type( 10, 5, "Press <enter> for the right side scroll:", 0, TERM_NORMAL );
	term_key();
	for( i = 0; i < t.num_cols; ++i ) {
		for( j = 0; j < t.num_rows; ++j ) {
			term_cur( j, 0 );
			term_insert_char( 1 );
			}
		term_flush();
		}
	}

void
attr_test() {
	int i = 3;

	term_clear( TERM_CLS_SCR );
	term_color( text_color );
	term_cur( i, 0 );
	term_printf( i++, 0, TERM_NORMAL, "Normal text, col=%d row=%d", t.col, t.row );
	term_type( i++, 0, "Bold text", 0, TERM_HILIGHT );
	term_type( i++, 0, "Inverse text", 0, TERM_INVERSE );
	term_type( i++, 0, "Blinking, inverse text", 0, TERM_INVERSE | TERM_BLINK );
	term_type( i++, 0, "Underlined text ( not on color screens )", 0, TERM_ULINE );
	term_type( i++, 0, "Bold, Inverse text", 0, TERM_HILIGHT | TERM_INVERSE );
	term_type( -1, -1, " (not on monochrome screens )", 0, TERM_NORMAL );
	term_attr_type( i++, 0, "A|lt|er|na|ting att|rib|utes", 0, 0, TERM_HILIGHT, '|' );
	}

void
term_ftest() {
	char buf[50];
	char def[50] = "test data";

	term_clear( TERM_CLS_SCR );
	term_type( 5, 0, "Term_field()=", 0, TERM_NORMAL );
	term_field( 5, 15, buf, 30, def, TERM_INVERSE );
	term_printf( 10, 0, TERM_NORMAL, "Input was |%s|", buf );
	}

void
color_test() {
	int f, b;

	term_clear( TERM_CLS_SCR );
	for ( f = 0; f < 8; ++f )
		for ( b = 0; b < 8; ++b )
			term_printf( f * 2, b * 2,
				(0x80 | b << 4 | f) << 8, "%1d%1d", f, b );
	}

void
speed_test() {
	time_t t1, t2, t3, t4;
	register int i, j, loop = 8000;

	term_clear( TERM_CLS_SCR );
	term_type( 0, 0, "Measuring term_flush() of full screen time...",
		0, TERM_NORMAL );
	time ( &t1 );
	for ( i = 0; i < 5000; ++i ) {
		t.region1 = 0;
		t.region2 = t.num_rows - 1;
		term_flush();
		}
	time ( &t2 );

	term_printf( t.num_rows-2, 0, TERM_NORMAL,
		"%2d seconds for 5000 term_flush() calls.", (int) (t2 - t1) );
	term_printf( t.num_rows-1, 0, TERM_NORMAL,
		"%2.04f seconds/screen, %4.02f milliseconds/screen.",
		(float) (t2-t1) / 5000.0,
		(float) ((t2-t1) / 5000.0)*1000 );
	term_key();

	term_clear( TERM_CLS_SCR );
	term_type( 0, 0, "Measuring term_flush() of single line time...",
		0, TERM_NORMAL );
	time ( &t1 );
	for ( i = 0; i < 5000; ++i ) {
		t.region1 = 0;
		t.region2 = 0;
		term_flush();
		}
	time ( &t2 );

	term_printf( t.num_rows-2, 0, TERM_NORMAL,
		"%2d seconds for 5000 term_flush() calls.", (int) (t2-t1) );
	term_printf( t.num_rows-1, 0, TERM_NORMAL,
		"%2.04f seconds/screen, %4.02f milliseconds/screen.",
		(float) (t2-t1) / 5000.0,
		(float) ((t2-t1) / 5000.0)*1000 );
	term_key();

	time( &t1 );
	for ( j = 0; j < 500; ++j ) {
		term_clear( TERM_CLS_SCR );
		term_color( text_color );
		term_box( 3, 3, 60, 20, TERM_HILIGHT );
		for ( i = 0; i < 15; ++i ) {
			term_type( i + 4, 10, "Testing:", 0, TERM_HILIGHT );
			}
		term_flush();
		}
	time( &t2 );
	term_printf( t.num_rows-2, 0, TERM_NORMAL, "%2d seconds for 500 screens.",
		(int) (t2-t1) );
	term_printf( t.num_rows-1, 0, TERM_NORMAL, "%2.04f seconds/screen, %4.02f milliseconds/screen.",
		(float) (t2-t1) / 500.0,
		(float) ((t2-t1) / 500.0)*1000 );
	term_key();

	term_clear( TERM_CLS_SCR );
	term_color( text_color );
	term_printf( 0, 0, TERM_NORMAL,
		"Update line 1 and 10 by repainting entire screen ( %d times )",
		loop );
	time( &t1 );
	for ( j = 0; j < loop; ++j ) {
		term_type( 1, 0, "Status line", 0, TERM_NORMAL );
		term_type( 10, 0, "Work line", 0, TERM_NORMAL );
		term_flush();
		}
	time( &t2 );

	term_clear( TERM_CLS_SCR );
	term_color( text_color );
	term_printf( 0, 0, TERM_NORMAL,
		"Update line 1 and 10 by repainting only lines 1 and 10 ( %d times )",
		loop );
	term_clear( TERM_CLS_EOL );
	time( &t3 );
	for ( j = 0; j < loop; ++j ) {
		term_type( 1, 0, "Status line", 0, TERM_NORMAL );
		term_flush();
		term_type( 10, 0, "Work line", 0, TERM_NORMAL );
		term_flush();
		}
	time( &t4 );

	term_printf( t.num_rows-2, 0, TERM_NORMAL,
		"%2d vs. %2d seconds for %d screens.",
		(int) (t2 - t1), (int) (t4 - t3), loop );
	term_printf( t.num_rows-1, 0, TERM_NORMAL,
	"%2.04f vs %2.04f seconds/screen, %4.02f vs. %4.02f milliseconds/screen",
		(float) (t2-t1) / (float) loop,
		(float) (t4-t3) / (float) loop,
		(float) ((t2-t1) / (float) loop)*1000,
		(float) ((t4-t3) / (float) loop)*1000 );
	term_key();
	}

void
prompt( char *msg ) {
	term_type( t.num_rows-1, 0, msg, 0, text_color );
	term_clear( TERM_CLS_EOS );
	term_key();
	}

void
in_test() {
	int c, i;

restart:	
	i = 0;
	term_clear( TERM_CLS_SCR );
	term_color( text_color );
	term_type( 0, 3, "term_key() / getch() Test Program", 0, TERM_HILIGHT );
	term_attr_type( 2, 0,
	"|Command keys:|  |^d|elay |^k|eypad |^h|alf_delay |^t|imeout_mode e|^x|it",
		0, 0, TERM_HILIGHT, '|' );
	term_type( 6, 0, "Function Key mask:", 0, TERM_HILIGHT );
	term_cur( 7, 0 );
	for( i = 0; i < 16; ++i )
		term_printf( -1, -1, TERM_NORMAL, "%04.4x ", __cur_term->keystart[ i ] );
	term_printf( 8, 0, TERM_NORMAL, "Holdoff delay: %u00 milliseconds",
		(print_rate != -1) ? print_rate : 3 );
	show_status();
	timeout_mode = ct->_time;
	for(;;) {
		show_status();
		term_type( 9, 0, "Press a key: ", 0, TERM_HILIGHT );
		term_clear( TERM_CLS_EOL );
		c = term_key();
		switch( c ) {
			case 0x04:		/* ^d for delay mode		*/
				delay_mode = !delay_mode;
				__nodelay( 0, delay_mode ? 1 : 0 );
				break;
			case 0x08:		/* ^h for half_delay mode	*/
				half_mode = half_mode ? 0 : 5;
				__halfdelay( half_mode );
				break;
			case 0x0b:		/* ^k for keypad mode		*/
				keypad_mode = !keypad_mode;
				__keypad( 0, keypad_mode ? 1 : 0 );
				break;
			case 0x14:		/* ^t for timeout mode		*/
				timeout_mode = !timeout_mode;
				__notimeout( 0, timeout_mode ? 0 : 1 );
				break;
			case K_RESIZE:
				goto restart;
			}
		++i;
		if ( term_delete_line( 10, 1 ) == -1 )
			term_type( t.num_rows-1, 0, "Delete line not supported.", 0, TERM_NORMAL );
		term_printf( t.num_rows - 1, 0, TERM_NORMAL,
						"%3d: c = %04x, %04o  %s", i, c, c, key_name( c ) );
		term_clear( TERM_CLS_EOL );
		if ( c == 0x18 ) return;		/* ^x to exit	*/
		}
	}

char *
key_name( unsigned c ) {
	static char name[20];
	unsigned c1;
	int dx, dy;

	name[0] = '\0';
	if (c >= ' '  &&  c <= 0x7f) {
		name[0] = c;
		name[1] = '\0';
		return( name );
		}
	if ( c == 0xffff ) {
		strcat( name, "Timeout" );
		return( name );
		}

	if ( (c1 = c & K_CLASS) == K_MOUSE_EVENT ) {
		if ( c1 == K_MOUSE_EVENT )	strcpy( name, "Mouse event:" );

		if ( c & K_MOUSE_XMASK == 0 ) 			dx = 0;
		else 			dx = 1 << ( ((c & K_MOUSE_XMASK) >> 4)-1 );
		if ( c & K_MOUSE_XDIR ) dx = -dx;
		if ( c & K_MOUSE_YMASK == 0 ) 			dy = 0;
		else 			dy = 1 << ( (c & K_MOUSE_YMASK)-1 );
		if ( c & K_MOUSE_YDIR ) dy = -dy;
		sprintf( name, "dx=%2d dy=%2d ", dx, dy );
		c1 = (c & 0x0f00) >> 8;
		strcat( name, ( c1 & (unsigned) _MOUSE_LEFT )	? "L" : " " );
		strcat( name, ( c1 & (unsigned) _MOUSE_MIDDLE )	? "M" : " " );
		strcat( name, ( c1 & (unsigned) _MOUSE_RIGHT )	? "R" : " " );
		}
	return( name );
	}

void
show_status() {
	int i;

	term_printf(  3,  0, TERM_HILIGHT, "Status: " );
	term_printf( -1, -1, TERM_NORMAL, "Delay mode: %s ", delay_mode ? "on " : "off" );
	term_printf( -1, -1, TERM_NORMAL, "Half delay: 0.%1usec ", half_mode );
	term_printf( -1, -1, TERM_NORMAL, "Keypad mode: %s ", keypad_mode ? "on " : "off" );
	term_printf( -1, -1, TERM_NORMAL, "Timeout mode: %s ", timeout_mode ? "on " : "off" );

	term_type( 5, 0, "Input queue: ", 0, TERM_HILIGHT );
	for( i = 0; i < 20; ++i )
		term_printf( -1, -1, TERM_NORMAL, " %02.2x", __cur_term->_input_queue[ i ] );
	}

/*---- tparm test routines ---------------------------------------------*/

int top;

char *
display( char *str ) {
	static char buf[100];
	char chname[ 10 ];
	register char *s = str;
	register char *d = buf;

	*d = '\0';
	while ( *s ) {
		if ( *s >= ' '  &&  *s <= 0x7f ) *d++ = *s;
		else {
			sprintf( chname, "{%02x}", *s );
			strcpy( d, chname );
			d += strlen( chname );
			}
		++s;
		}
	*d = '\0';
	return( buf );
	}

void
test( char *fmt, int p1, int p2 ) {
	char *p;

	p = __tparm( fmt, p1, p2 );
	term_printf( top, 0, TERM_NORMAL, "%s =", display( fmt ) );
	term_printf( -1, -1, TERM_NORMAL, "%s", display( p ) );
	++top;
	}

void
tparm_test() {
	char *p, *f;

	term_clear( TERM_CLS_SCR );
	top = 0;
	test( cursor_address, 10, 20 );
	test( "Double Percent %%test%%", 0, 0 );
	test( "Register 'a' %p1%Patest%ga%d", 120, 0 );
	test( "text %p1%5.2dtest", 3, 0 );
	test( "text %p1%:+5.2dtest", 4, 0 );
	test( "text %p1%:-5.2dtest", 5, 0 );
	p = __tparm( f = "text %p1%:-5stest", "the", 0 );
	term_printf( top++, 0, 0, "%s =", display( f ) );
	term_printf( -1, -1, 0, "%s", display( p ) );
	p = __tparm( f = "Strlen(\"testing\") %p1%l%d", "testing", 0 );
	term_printf( top++, 0, 0, "%s =", display( f ) );
	term_printf( -1, -1, 0, "%s", display( p ) );
	test( "Decimal constant ab%{123}%dcd", 0, 0 );
	test( "10+5= %{5}%p1%+%d...", 10, 0 );
	test( "15-5= %{15}%p1%-%d...", 5, 0 );
	test( "12*3= %p1%{3}%*%d...", 12, 0 );
	test( "12/4= %p1%{4}%/%d...", 12, 0 );
	test( "12%%5= %p1%{5}%m%d...", 12, 0 );
	term_type( t.num_rows-1, 0, "Press <Enter> to continue...", 0, 0 );
	term_key();
	term_clear( TERM_CLS_SCR );
	top = 0;
	term_printf( top++, 0, 0, "if 2<5 then display 10: " );
	test( "%?%p1%{5}%<%t%p2%d", 2, 10 );
	test( "%?%p1%{5}%<%t%p2%d%;test", 2, 10 );
	term_printf( top++, 0, 0, "if 7<5 then display 10: " );
	test( "%?%p1%{5}%<%t%p2%d", 7, 10 );
	test( "%?%p1%{5}%<%t%p2%d%;test", 7, 10 );
	term_printf( top++, 0, 0, "if 2<5 display 10 else 7 : " );
	test( "%?%p1%{5}%<%t%p2%d%e%{7}%d", 2, 10 );
	test( "%?%p1%{5}%<%t%p2%d%e%{7}%d%;test", 2, 10 );
	term_printf( top++, 0, 0, "if 7<5 display 10 else 7 : " );
	test( "%?%p1%{5}%<%t%p2%d%e%{7}%d", 7, 10 );
	test( "%?%p1%{5}%<%t%p2%d%e%{7}%d%;test", 7, 10 );
	test( cursor_address, 10, 20 );
	term_type( t.num_rows-1, 0, "Press <Enter> to continue...", 0, 0 );
	term_key();
	}

/*---- Pop-up menu test routines -------------------------------------------*/

#define	MAX_SCREENS		5		/* Maximum # of screens that can be saved	*/

int bytes = 0;						/* # of bytes / saved character			*/
int top_screen = 0;					/* Number of top window					*/
struct screen_save *segments[ MAX_SCREENS ];	/* Buffer of screen segments*/

char *a_menu[] = { "Axxx ", "Bxxx ", "Cxxx ", "Dxxx ", "Exxx ", "Quit ", NULL };
char *b_menu[] = { "Ayyy ", "Byyy ", "Cyyy ", "Dyyy ", "Eyyy ", "Quit ", NULL };
char *c_menu[] = { "Azzz ", "Bzzz ", "Czzz ", "Dzzz ", "Ezzz ", "Quit ", NULL };

struct screen_save {				/* Structure to contain saved windows	*/
			int		cursor_x, cursor_y;
			int		x, y, width, length;
			char	data[1];
			};

void
pop_test() {
	register char *p2;

	term_clear( TERM_CLS_SCR );		/* Set screen buffer to a known state	*/

	pop_up( 3, 9, 20, 8, TERM_BLUE | TERM_CYAN_BG | TERM_HILIGHT );
	p2 = a_menu[0];
	for(;;) {
		p2 = term_lmenu( 4, 11, a_menu, p2, TERM_BLUE | TERM_CYAN_BG, NULL, 3 );
		if ( *p2 == 'Q' ) break;
		else submenu1( 5 + *p2 - 'A' );
		}
	pop_down();
	}

void
submenu1( int row ) {
	register char *p2;

	pop_up( row, 13, 20, 8, TERM_RED | TERM_YELLOW_BG | TERM_HILIGHT );
	p2 = b_menu[0];
	for(;;) {
		p2 = term_lmenu( row + 1, 15, b_menu, p2, TERM_RED | TERM_YELLOW_BG, NULL, 3 );
		if ( *p2 == 'Q' ) break;                                                       	
		else submenu2( row + *p2 - 'A' + 2 );
		}
	pop_down();
	}

void
submenu2( int row ) {
	register char *p2;

	pop_up( row, 18, 20, 8, TERM_BLUE | TERM_WHITE_BG | TERM_HILIGHT );
	p2 = c_menu[0];
	for(;;) {
		p2 = term_lmenu( row+1, 20, c_menu, p2, TERM_BLUE | TERM_WHITE_BG, NULL, 3 );
		if ( *p2 == 'Q' ) break;
		}
	pop_down();
	}

void
pop_up( int y, int x, int width, int length, unsigned attr ) {
	register unsigned			i;
	register struct screen_save	*s;
	register char				*p;
	unsigned					offset;

	if ( !bytes ) {
		bytes = term_save_image( 0, 0, 0, 0 );
		__memsetw( segments, 0, MAX_SCREENS );
		}
	offset = width * bytes;
	if ( top_screen >= MAX_SCREENS ) {
		fprintf( stderr, "pop: Not enough screen buffers.\n" );
		exit( -1 );
		}
	if ( !( s = segments[ top_screen ] =
		malloc( offset*length + sizeof(struct screen_save)+15) )) {
		fprintf( stderr, "pop: Out of memory\n" );
		exit( -1 );
		}
	s->cursor_x = term_state.col;
	s->cursor_y = term_state.row;
	s->x = x;	 	s->length	= length;
	s->y = y;	 	s->width	= width;
	p = s->data;
	for ( i = y; i < length + y; ++i ) {	/* Save each line of the window	*/
		term_save_image( i, x, p, width );
		p += offset;
		}
	++top_screen;
	term_box_fill( y, x, width, length, attr, TERM_BOX_DBL_TOP, ' ' );
	term_color( attr );
	}

void
pop_down() {
	register int				i;
	register struct screen_save	*s;
	register char				*p;
	unsigned x, y, width, length, offset;

	if ( --top_screen < 0 ) {
		fprintf( stderr, "pop: Restored too many screens\n" );
		exit( -1 );
		}
	if ( !( s = segments[ top_screen ] ) ) {
		fprintf( stderr, "pop: Attempt to restore unused screen buffer.\n" );
		exit( -1 );
		}
	p = s->data;
	x = s->x;			 	length	= s->length; 
	y = s->y;			 	width	= s->width;
	offset = width * bytes;
	for ( i = y; i < y + length; ++i ) { /* Restore each line of the window	*/
		term_restore_image( i, x, p, width );
		p += offset;
		}
	term_cur( s->cursor_y, s->cursor_x );
	free( s );
	segments[ top_screen ] = NULL;
	}

void
do_shell() {
	term_cur( t.num_rows - 1, 0 );
	term_restore();
	printf( "\nType 'exit' to return to termtest.\n" );
	fflush( stdout );
	system( "/bin/qsh" );
	term_init();
	}

void
mouse_test() {
	unsigned c;
	int mx, my, cx, cy, dx, dy, off;

	if ( !mouse ) return;

resized:
	mx = my = 10;
	cx = cy = 20;
	dx = dy = 22;
	term_clear( TERM_CLS_SCR );
	term_type( t.num_rows-1, 0,
"Arrow keys to move the 'c', mouse to move 'm', drag to move 'd', 'x' to exit",
		0, TERM_NORMAL );
	term_type( cy, cx, "c", 1, TERM_NORMAL );
	term_type( my, mx, "m", 1, TERM_NORMAL );
	term_type( dy, dx, "d", 1, TERM_NORMAL );
	for(;;) {
		c = term_key();
		if ( (c & K_CLASS) == K_MOUSE_EVENT ) {
			if ( c & K_MOUSE_XMASK ) {
				if ( c & K_MOUSE_XMASK == 0 )
					off = 0;
				else
					off = 1 << ( ((c & K_MOUSE_XMASK) >> 4)-1 );
				if ( c & K_MOUSE_XDIR ) off = -off;

				if ( c & K_MOUSE_BLEFT ) {
					term_type( dy, dx, " ", 1, TERM_NORMAL );
					dx += off;
					if ( dx > t.num_cols-1 ) dx = t.num_cols-1;
					if ( dx < 0 ) dx = 0;
					term_type( dy, dx, "d", 1, TERM_NORMAL );
					}
				else {
					term_type( my, mx, " ", 1, TERM_NORMAL );
					mx += off;
					if ( mx > t.num_cols-1 ) mx = t.num_cols-1;
					if ( mx < 0 ) mx = 0;
					term_type( my, mx, "m", 1, TERM_NORMAL );
					}
				}
			if ( c & K_MOUSE_YMASK ) {
				if ( c & K_MOUSE_YMASK == 0 )
					off = 0;
				else
					off = 1 << ( (c & K_MOUSE_YMASK)-1 );
				if ( c & K_MOUSE_YDIR ) off = -off;

				if ( c & K_MOUSE_BLEFT ) {
					term_type( dy, dx, " ", 1, TERM_NORMAL );
					dy -= off;
					if ( dy > t.num_rows-1 ) dy = t.num_rows-1;
					if ( dy < 0 ) dy = 0;
					term_type( dy, dx, "d", 1, TERM_NORMAL );
					}
				else {
					term_type( my, mx, " ", 1, TERM_NORMAL );
					my -= off;
					if ( my > t.num_rows-1 ) my = t.num_rows-1;
					if ( my < 0 ) my = 0;
					term_type( my, mx, "m", 1, TERM_NORMAL );
					}
				}
			}
		else {
			term_type( cy, cx, " ", 1, TERM_NORMAL );
			switch( c ) {
				case K_UP:		if ( --cy < 0 ) cy = 0;
								break;
				case K_DOWN:	if ( ++cy > t.num_rows-1 ) cy = t.num_rows-1;
								break;
				case K_LEFT:	if ( --cx < 0 ) cx = 0;
								break;
				case K_RIGHT:	if ( ++cx > t.num_cols-1 ) cx = t.num_cols-1;
								break;
	   			case K_RESIZE:	goto resized;
				case 'x':		return;
				}
			term_type( cy, cx, "c", 1, TERM_NORMAL );
			}
		}
	}
